@@ -6,7 +6,10 @@ import io
 from . import sobjects
 from lht.util import field_types
 from lht.util import stage
+from lht.util import table_creator
 import os
+from lht.util import merge
+
 
 def create_batch_query(access_info, query):
 	"""Creates a batch query job in Salesforce using the Bulk Query API.
@@ -74,6 +77,7 @@ def query_status(access_info, job_type, job_id):
 			query_statuses.append(results.json())
 			break
 		if results.status_code > 200:
+			print(results.status_code)
 			print("not logged in")
 			exit(0)
 		records = len(results.json()['records'])
@@ -151,11 +155,14 @@ def get_query_ids(access_info):
 
 	return jobs
 
-def get_bulk_results_direct(session, access_info, job_id, sobject, schema, table, database=None):
+def get_bulk_results_direct(session, access_info, job_id, sobject, schema, table, snowflake_fields=None, database=None, force_full_sync=False):
+	print(f"🔍 get_bulk_results_direct called with force_full_sync={force_full_sync}")
+	print(f"🔍 force_full_sync type: {type(force_full_sync)}")
+	print(f"🔍 force_full_sync value: {repr(force_full_sync)}")
+	
 	# Auto-detect database if not provided
 	if database is None:
 		database = session.sql('SELECT CURRENT_DATABASE()').collect()[0][0]
-		print(f"🔍 DEBUG: Auto-detected database: {database}")
 	"""Fetches and processes bulk query results from Salesforce, loading them directly into a Snowflake table.
 
 	Args:
@@ -187,86 +194,50 @@ def get_bulk_results_direct(session, access_info, job_id, sobject, schema, table
 		print('The job is not ready.  Retry in a few minutes')
 		return None
 	
-	# Get field descriptions for formatting
-	query_string, df_fields = sobjects.describe(access_info, sobject)
+	# Always get both field types from describe to ensure we have the correct information
+	query_string, df_fields, snowflake_fields = sobjects.describe(access_info, sobject)
 	
 	# Process first batch
 	csv_content = results.text
 	print(f"\n\nPROCESSING BATCH 1")
 	
 	# Load and process data directly (no stage upload needed)
-	df = pd.read_csv(io.StringIO(csv_content))
-	formatted_df = field_types.format_sync_file(df, df_fields)
-	formatted_df = formatted_df.replace(np.nan, None)
+	# CRITICAL: Force string reading to prevent pandas from converting numeric strings to floats
+	# This prevents "20" from becoming 20.0 and then "20.0"
+	df = pd.read_csv(io.StringIO(csv_content), dtype=str)
 	
 	# Set the current database and schema context
 	session.sql(f"USE DATABASE {database}").collect()
 	session.sql(f"USE SCHEMA {schema}").collect()
 	
-	# Use fully qualified table name for write_pandas
-	fully_qualified_table = f"{schema}.{table}"
-	
-	# CRITICAL FIX: Create table with correct schema BEFORE write_pandas
+	# Use centralized table creation utility
 	try:
-		print(f"🔍 DEBUG: About to create table {schema}.{table}")
-		print(f"🔍 DEBUG: df_fields contains {len(df_fields)} fields: {list(df_fields.keys())}")
-		
-		# Check if table already exists first
-		print(f"🔍 DEBUG: Checking if table already exists...")
-		try:
-			table_check = session.sql(f'SHOW TABLES IN SCHEMA "{schema}"').collect()
-			table_names = [row['name'] for row in table_check]
-			print(f"🔍 DEBUG: Tables in schema {schema}: {table_names}")
-			print(f"🔍 DEBUG: Looking for table: {table}")
-			
-			if table in table_names:
-				print(f"🔍 DEBUG: Table {schema}.{table} already exists, skipping creation")
-			else:
-				print(f"🔍 DEBUG: Table {schema}.{table} does not exist, creating it...")
-				create_table_sql = _build_create_table_sql(schema, table, df_fields)
-				print(f"🔍 DEBUG: CREATE TABLE SQL: {create_table_sql}")
+		print(f"🚀 Creating table {schema}.{table}...")
+		table_creator.ensure_table_exists_for_dataframe(
+			session=session,
+			schema=schema,
+			table=table,
+			df_fields=df_fields,
+			snowflake_fields=snowflake_fields,
+			force_full_sync=force_full_sync,
+			database=database
+		)
+		print(f"✅ Table creation completed successfully")
 				
-				# Execute CREATE TABLE statement
-				print(f"🔍 DEBUG: Executing CREATE TABLE statement...")
-				result = session.sql(create_table_sql).collect()
-				print(f"✅ Table created with correct schema")
-				
-				# Verify table was created
-				print(f"🔍 DEBUG: Verifying table creation...")
-				table_check_after = session.sql(f'SHOW TABLES IN SCHEMA "{schema}"').collect()
-				table_names_after = [row['name'] for row in table_check_after]
-				print(f"🔍 DEBUG: Tables after creation: {table_names_after}")
-				
-				if table not in table_names_after:
-					raise Exception(f"Table {schema}.{table} was not created successfully")
-				else:
-					print(f"🔍 DEBUG: Table {schema}.{table} verified successfully")
-		
-		except Exception as table_check_error:
-			print(f"🔍 DEBUG: Error checking/creating table: {table_check_error}")
-			print(f"🔍 DEBUG: Falling back to auto_create_table=True")
-			# Fallback: let Snowflake create the table automatically
-			session.write_pandas(formatted_df, fully_qualified_table, quote_identifiers=False, auto_create_table=True, overwrite=False, use_logical_type=False, on_error="CONTINUE")
-			print(f"✅ Data loaded with auto-created table")
-			return results
-		
-		# Now load data into existing table (no auto-create, no overwrite)
-		print(f"🔍 DEBUG: Loading data into table {fully_qualified_table}")
-		session.write_pandas(formatted_df, fully_qualified_table, quote_identifiers=False, auto_create_table=False, overwrite=False, use_logical_type=False, on_error="CONTINUE")
-		print(f"✅ Data loaded successfully")
-		
+		df_str = df.astype(str)
+		df = None
+		print(f"📊 Processing first batch of data")
+		session.write_pandas(df_str, schema=schema, table_name="tmp_"+table, auto_create_table=True, overwrite=True, quote_identifiers=False, table_type="temporary")
+		df_str = None
+		transformed_data = merge.transform_and_match_datatypes(session, "tmp_"+table, table)
+		session.sql(f"Insert into {table} select {transformed_data} from tmp_{table}").collect()
+		print(f"✅ First batch loaded successfully")
 	except Exception as e:
 		print(f"❌ Failed to create table or load data: {e}")
-		print(f"🔍 DEBUG: Final fallback - trying auto_create_table=True")
-		try:
-			# Ultimate fallback: let Snowflake handle everything
-			session.write_pandas(formatted_df, fully_qualified_table, quote_identifiers=False, auto_create_table=True, overwrite=False, use_logical_type=False, on_error="CONTINUE")
-			print(f"✅ Data loaded with ultimate fallback")
-		except Exception as final_error:
-			print(f"❌ Ultimate fallback also failed: {final_error}")
-			raise final_error
+		raise Exception(f"Failed to load data into table {schema}.{table}: {e}")
 	
 	# Process remaining batches
+
 	counter = 2
 	while True:
 		if 'Sforce-Locator' not in results.headers:			
@@ -279,29 +250,24 @@ def get_bulk_results_direct(session, access_info, job_id, sobject, schema, table
 		csv_content = results.text
 		print(f"\n\nPROCESSING BATCH {counter}")
 		
-		df = pd.read_csv(io.StringIO(csv_content))
-		formatted_df = field_types.format_sync_file(df, df_fields)
-		formatted_df = formatted_df.replace(np.nan, None)
-	
-		# Set the current database and schema context
-		session.sql(f"USE DATABASE {database}").collect()
-		session.sql(f"USE SCHEMA {schema}").collect()
-		
-		# Use fully qualified table name for write_pandas
-		fully_qualified_table = f"{schema}.{table}"
-		
-		try:
-			session.write_pandas(formatted_df, fully_qualified_table, quote_identifiers=False, auto_create_table=False, overwrite=False, use_logical_type=True, on_error="CONTINUE")
-			print(f"✅ Batch {counter} loaded successfully")
-		except Exception as e:
-			print(f"❌ Batch {counter} failed: {e}")
-			raise
+		# CRITICAL: Force string reading to prevent pandas from converting numeric strings to floats
+		# This prevents "20" from becoming 20.0 and then "20.0"
+		df = pd.read_csv(io.StringIO(csv_content), dtype=str)
 		
+
+		df_str = df.astype(str)
+		df = None
+		print(f"📊 Processing batch {counter}")
+		session.write_pandas(df_str, schema=schema, table_name="tmp_"+table, auto_create_table=True, overwrite=True, quote_identifiers=False, table_type="temporary")
+		transformed_data = merge.transform_and_match_datatypes(session, "tmp_"+table, table)
+		session.sql(f"Insert into {table} select {transformed_data} from tmp_{table}").collect()
+		print(f"✅ Batch {counter} loaded successfully using save_as_table")
+		df_str = None
 		counter += 1
 	
 	return results
 
-def get_bulk_results(session, access_info, job_id, sobject, schema, table, use_stage=False, stage_name=None, database=None):
+def get_bulk_results(session, access_info, job_id, sobject, schema, table, snowflake_fields=None, use_stage=False, stage_name=None, database=None, force_full_sync=False):
 	"""Fetches and processes bulk query results from Salesforce, loading them into a Snowflake table.
 	
 	This function now uses direct DataFrame-to-table loading for optimal performance.
@@ -326,7 +292,10 @@ def get_bulk_results(session, access_info, job_id, sobject, schema, table, use_s
 		pandas.errors.EmptyDataError: If the CSV data is empty or malformed.
 		snowflake.snowpark.exceptions.SnowparkSQLException: If Snowflake write operation fails.
 	"""
-	return get_bulk_results_direct(session, access_info, job_id, sobject, schema, table, database)
+	print(f"🔍 get_bulk_results called with force_full_sync={force_full_sync}")
+	print(f"🔍 force_full_sync type: {type(force_full_sync)}")
+	print(f"🔍 force_full_sync value: {repr(force_full_sync)}")
+	return get_bulk_results_direct(session, access_info, job_id, sobject, schema, table, snowflake_fields, database, force_full_sync)
 
 def delete_query(access_info, job_id):
 	"""Deletes a Salesforce query job by ID using the Bulk Query API.
@@ -405,49 +374,10 @@ def test_snowflake_permissions(session, schema, table, database=None):
 	# Auto-detect database if not provided
 	if database is None:
 		database = session.sql('SELECT CURRENT_DATABASE()').collect()[0][0]
-		print(f"🔍 DEBUG: Auto-detected database: {database}")
 	
-	print(f"\n🔍 DEBUG: Testing Snowflake permissions and context")
-	print(f"🔍 DEBUG: Target schema: {schema}")
-	print(f"🔍 DEBUG: Target table: {table}")
-	print(f"🔍 DEBUG: Target database: {database}")
-	
-	# Check current context
-	current_db = session.sql('SELECT CURRENT_DATABASE()').collect()[0][0]
-	current_schema = session.sql('SELECT CURRENT_SCHEMA()').collect()[0][0]
-	current_role = session.sql('SELECT CURRENT_ROLE()').collect()[0][0]
-	
-	print(f"🔍 DEBUG: Current database: {current_db}")
-	print(f"🔍 DEBUG: Current schema: {current_schema}")
-	print(f"🔍 DEBUG: Current role: {current_role}")
-	
-	# Test database permissions
-	try:
-		print(f"🔍 DEBUG: Testing database access: {database}")
-		session.sql(f"SHOW SCHEMAS IN DATABASE {database}").collect()
-		print(f"✅ DEBUG: Database access successful")
-	except Exception as e:
-		print(f"❌ DEBUG: Database access failed: {e}")
-	
-	# Test schema permissions
-	try:
-		print(f"🔍 DEBUG: Testing schema access: {database}.{schema}")
-		session.sql(f"SHOW TABLES IN SCHEMA {database}.{schema}").collect()
-		print(f"✅ DEBUG: Schema access successful")
-	except Exception as e:
-		print(f"❌ DEBUG: Schema access failed: {e}")
-	
-	# Test table creation permissions
-	try:
-		print(f"🔍 DEBUG: Testing table creation in schema: {database}.{schema}")
-		test_table = f"{database}.{schema}.TEST_PERMISSIONS_TEMP"
-		session.sql(f"CREATE OR REPLACE TEMPORARY TABLE {test_table} (test_col STRING)").collect()
-		session.sql(f"DROP TABLE {test_table}").collect()
-		print(f"✅ DEBUG: Table creation test successful")
-	except Exception as e:
-		print(f"❌ DEBUG: Table creation test failed: {e}")
-	
-	print(f"🔍 DEBUG: Permission test completed\n")
+	# Set the current database and schema context
+	session.sql(f"USE DATABASE {database}").collect()
+	session.sql(f"USE SCHEMA {schema}").collect()
 
 def cleanup_completed_jobs(access_info, max_age_hours=24):
 	"""Deletes completed Salesforce Bulk API 2.0 jobs that are older than the specified age.
@@ -596,44 +526,3 @@ def delete_specific_job(access_info, job_id):
 			'error': str(e)
 		}
 
-def _build_create_table_sql(schema: str, table: str, df_fields: dict) -> str:
-	"""
-	Build CREATE TABLE SQL statement based on Salesforce field definitions.
-	
-	Args:
-		schema: Snowflake schema name
-		table: Snowflake table name
-		df_fields: Dictionary mapping field names to pandas dtypes
-		
-	Returns:
-		str: CREATE TABLE SQL statement
-	"""
-	# Map pandas dtypes to Snowflake types
-	type_mapping = {
-		'object': 'VARCHAR(16777216)',      # String fields (including PO_Number__c)
-		'int64': 'NUMBER(38,0)',            # Integer fields
-		'float64': 'DOUBLE',                 # Float fields
-		'bool': 'BOOLEAN',                   # Boolean fields
-		'datetime64': 'TIMESTAMP_NTZ',       # Datetime fields
-		'datetime64[ns]': 'TIMESTAMP_NTZ',   # Datetime fields with nanoseconds
-	}
-	
-	# Build column definitions
-	columns = []
-	for field_name, dtype in df_fields.items():
-		# Convert field name to uppercase to match DataFrame
-		field_upper = field_name.upper()
-		
-		# Map the dtype to Snowflake type
-		snowflake_type = type_mapping.get(dtype, 'VARCHAR(16777216)')  # Default to VARCHAR
-		
-		# Add column definition
-		columns.append(f'"{field_upper}" {snowflake_type}')
-	
-	# Build CREATE TABLE statement
-	column_defs = ',\n\t'.join(columns)
-	create_sql = f"""CREATE OR REPLACE TABLE "{schema}"."{table}" (
-	{column_defs}
-)"""
-	
-	return create_sql
\ No newline at end of file
