@@ -23,6 +23,7 @@ def describe(access_info, sobject, lmd=None):
 	create_table_fields = ''
 	cfields = []
 	df_fields = {}
+	snowflake_fields = {}  # For table creation with proper Snowflake types
 
 	if results.status_code > 200:
 		print("you are not logged in")
@@ -35,7 +36,16 @@ def describe(access_info, sobject, lmd=None):
 		# Skip compound fields
 		if row['name'] in cfields:
 			continue
-
+		
+		# Skip fields the user doesn't have access to
+		if not row.get('accessible', True):
+			print(f"⚠️ Skipping inaccessible field: {row['name']}")
+			continue
+		
+		# Skip fields that can't be retrieved
+		if not row.get('retrieveable', True):
+			print(f"⚠️ Skipping non-retrievable field: {row['name']}")
+			continue
 		
 		if len(query_fields) == 0:
 			pass
@@ -44,7 +54,15 @@ def describe(access_info, sobject, lmd=None):
 			
 		query_fields += row['name']
 		df_fields[row['name']] = field_types.df_field_type(row)
+		snowflake_fields[row['name']] = field_types.salesforce_field_type(row)
 	query_string = "select+"+query_fields+"+from+{}".format(sobject)
 	if lmd is not None:
 		query_string = query_string + "+where+LastModifiedDate+>+{}".format(lmd)
-	return query_string, df_fields
\ No newline at end of file
+	
+	# Returning field descriptions from Salesforce
+	print(f"  - df_fields keys: {list(df_fields.keys())}")
+	print(f"  - df_fields values: {list(df_fields.values())}")
+	print(f"  - snowflake_fields keys: {list(snowflake_fields.keys())}")
+	print(f"  - snowflake_fields values: {list(snowflake_fields.values())}")
+	
+	return query_string, df_fields, snowflake_fields
\ No newline at end of file
